#include "util.h"
#include "array.h"
#include "st.h" 

#include "bdd.h"
#include "bdd_int.h"
#include "bdd_bf.h"
#include "bdd_bf_int.h"


int 
bdd_bf_bdd_node_compare(obj1, obj2)
char *obj1;
char *obj2;
{
  return *((bdd_node **) obj1) - *((bdd_node **) obj2);
}

/*
 * 
 */
static bdd_node *
bdd_bf_and_multi_rec(mgr, operands)
bdd_manager *mgr;
array_t *operands;                    /* bdd_node * */
{
    bfbdd_info_t *bfbdd_info;
    bdd_safeframe frame;
    bdd_safenode then_result, else_result, result, var;
    bdd_safenode safe_operands, safe_else_nodes;
    bdd_variableId top_id;
    int i, num_operands, num_distinct_non_constants;
    array_t *then_nodes, *else_nodes;  /* bdd_node * */
    bdd_node *oprnd, *distinct_non_constant, *previous_oprnd;

    bfbdd_info = (bfbdd_info_t *)mgr->undef1;
    bfbdd_info->stats.bf_ops.calls++;
    num_operands = array_n(operands);

    /*
     * Sort the operands array based on the bdd_node addresses.  This increases the chance
     * of a cache hit, and it helps us to detect if two operands are complements of eachother
     * (because a complemented pointer just has the last bit of the address set, complemented
     * operands must be adjacent in the sorted array). NOTE: must sort after garbage collection also.
     */
    array_sort(operands, bdd_bf_bdd_node_compare); 

    /*
     * Check if the value of operands represents a terminal case.  We could reduce the number
     * of lines of code in this if/else by pushing the test of fml_type deeper, but at the
     * expense of slower code.
     *
     * Also, determine the minimum of the top variable IDs of the operands.  Assume BDD_ONE_ID acts like 
     * plus infinity.  If we hit a terminal case and return, the value of top_id is inconsequential.
     * We assume that no garbage collection can take place during this block of code.
     *
     * If we do hit a terminal case, the result does not go into the cache.  Thus, we will never again
     * have the chance to free operands: thus, free it now.
     */
    num_distinct_non_constants = 0;
    top_id = (bdd_variableId) BDD_ONE_ID;

    for (i = 0; i < num_operands; i++) {
	oprnd = array_fetch(bdd_node *, operands, i);  
	if (oprnd == BDD_ZERO(mgr)) {
	    bfbdd_info->stats.bf_ops.returns.trivial++;
	    array_free(operands);
	    return (BDD_ZERO(mgr));
	} else if (oprnd != BDD_ONE(mgr)) {
	    /* operand is not a constant */
	    top_id = MIN(top_id, BDD_REGULAR(oprnd)->id); /* only need to check non-constants */
	    if (num_distinct_non_constants == 0) {
		num_distinct_non_constants = 1;
		distinct_non_constant = oprnd;
		previous_oprnd = oprnd;
	    } else if (num_distinct_non_constants == 1) {
		if (oprnd != distinct_non_constant) {
		    num_distinct_non_constants = 2;
		    if (oprnd == BDD_NOT(previous_oprnd)) {
			bfbdd_info->stats.bf_ops.returns.trivial++;
			array_free(operands);
			return (BDD_ZERO(mgr));
		    } else {
			previous_oprnd = oprnd;
		    }
		} /* else: so far all non-constant operands are the same */
	    } else { 
		/* the only terminal conditions left to check are complements or ZERO */
		if (oprnd == BDD_NOT(previous_oprnd)) {
		    bfbdd_info->stats.bf_ops.returns.trivial++;
		    array_free(operands);
		    return (BDD_ZERO(mgr));
		} else {
		    previous_oprnd = oprnd;
		}
	    } 
	} /* else: is ONE */
    }

    if (num_distinct_non_constants == 0) {
	/* all operands are ONES */
	bfbdd_info->stats.bf_ops.returns.trivial++;
	array_free(operands);
	return (BDD_ONE(mgr));
    } else if (num_distinct_non_constants == 1) {
	/* there was only one distinct non-constant - return it */
	bfbdd_info->stats.bf_ops.returns.trivial++;
	array_free(operands);
	return (distinct_non_constant);
    } /* else: no ZEROs, 0 or more ONES, 2 or more distinct non-constants */

    /*
     * If we have already seen the same array of operands, then we are done with this 
     * branch of the recursion.  Note that use of the variable "oprnd" is only for efficiency
     * purposes - conceptually, this should be named result.  TODO: sort array by address first, 
     * to increase hit rate; can do this since AND is associative and commutative.
     */
    if (bdd_bf_cache_lookup(mgr, operands, &oprnd)) {
	bfbdd_info->stats.bf_ops.returns.cached++;
	array_free(operands);
	return oprnd;
    }

    /*
     * Based on the value of top_id and the oprnds_cof_info, apply Shannon expansion to the nodes 
     * in operands.
     */
    then_nodes = array_alloc(bdd_node *, num_operands);
    else_nodes = array_alloc(bdd_node *, num_operands);
    bdd_get_branches_array(top_id, operands, then_nodes, else_nodes, NIL(array_t));

    /*
     * Start the safeframe.  Normally, this is the first block of code of such a function. 
     * However, if we are CAREFUL, we see that nothing needs to be protected before this point.
     */
    bdd_safeframe_start(mgr, frame);
    bdd_safenode_declare(mgr, then_result);
    bdd_safenode_declare(mgr, else_result);
    bdd_safenode_declare(mgr, result);
    bdd_safenode_declare(mgr, var);
    bdd_bf_safenode_array(mgr, safe_operands, operands);

    /*
     * The garbage collector doesn't allow a node to be protected more than once.  Since this function
     * recursively protects its input, operands, we have to be careful how we treat the arrays then_nodes
     * and else_nodes.  It's assumed that from the point when then_nodes is set (in bdd_get_branches_array)
     * to the point it is read (in the call to bdd_bf_and_multi_rec), a garbage collection can't be
     * triggered.  However, the else_nodes array is different.  It must be protected during the first call
     * to bdd_bf_and_multi_rec, and then unprotected before its call to bdd_bf_and_multi_rec.
     * NOTE: because of this, the order of calls to bdd_bf_and_multi_rec is important.
     */
    bdd_bf_safenode_array(mgr, safe_else_nodes, else_nodes);  
    then_result.node = bdd_bf_and_multi_rec(mgr, then_nodes);
    safe_else_nodes.array = NIL(array_t);
    else_result.node = bdd_bf_and_multi_rec(mgr, else_nodes);

    /*
     * It is tempting to try and free the then_nodes and else_nodes arrays.  However, they have been
     * inserted into the bf_cache, so freeing them will cause bad behavior.  These arrays are 
     * eventually freed by bdd_bf_cache_reset.
     */

    /* 
     * Combine the results of the recursion down the then and else branches.  Note that this is the only
     * location where a BDD node can be created in mgr.  NOTE: since this call to bdd__ITE_ is always 
     * "trivial", the call to bdd__ITE_ could be inlined here and some of the irrelevant checks it
     * makes could be eliminated, to improve efficiency of this code.
     */
    var.node = bdd_find_or_add(mgr, top_id, BDD_ONE(mgr), BDD_ZERO(mgr));
    result.node = bdd__ITE_(mgr, var.node, then_result.node, else_result.node);

    /*
     * Insert the result into the cache.
     */
    bdd_bf_cache_insert(mgr, operands, result.node);  /* TODO: sort first */
    bfbdd_info->stats.bf_ops.returns.full++;

    /*
     * Clean up, and return.
     */
    bdd_safeframe_end(mgr);
    return (result.node);
}

/*
 * Set up the recursion for doing the multi-way AND or OR.  Convert a multi-way
 * OR into a multi-way AND using DeMorgan's Law.
 */
bdd_t *
bdd_bf_and_or_multi(operands, fml_type)
array_t *operands;
bdd_fml_type_t fml_type;
{
    bfbdd_info_t *bfbdd_info;
    bdd_t *oprnd, *result_bdd;
    bdd_manager *mgr;
    int i, num_operands;
    array_t *operand_nodes;    /* bdd_node * */
    bdd_node *result, *oprnd_node;
    

    /* 
     * Get the manager via the manager field of any bdd_t.  Set the sorting flag.
     */
    oprnd = array_fetch(bdd_t *, operands, 0);
    mgr = bdd_get_manager(oprnd);
    bfbdd_info = (bfbdd_info_t *)mgr->undef1;
    bfbdd_info->sort_cache_at_gc = TRUE;  /* this operation requires sorting the arrays */

    /*
     * Check once and for all, before recursion, that the fml_type is AND or OR.
     */
    if ((fml_type != BDD_FML_AND) && (fml_type != BDD_FML_OR)) {
	fail("bdd_and_or_multi: unknown formula type");
    }

    /* 
     * Convert operands into an array of bdd_nodes for recursion. If OR, first
     * complement all the operands.
     */
    num_operands = array_n(operands);
    operand_nodes = array_alloc(bdd_node *, num_operands);
    for (i = 0; i < num_operands; i++) {
	oprnd = array_fetch(bdd_t *, operands, i);
	if (fml_type == BDD_FML_AND) {
	    oprnd_node = oprnd->node;
	} else { /* fml_type == BDD_FML_OR */
	    oprnd_node = BDD_NOT(oprnd->node);
	}
	array_insert(bdd_node *, operand_nodes, i, oprnd_node);
    }

    /*
     * Build the BDD for the result.  Note that we are not protecting (using safenode, etc.) the variable
     * "result".  It's assumed that from the point that result is set, to the point where it is used
     * (inside bdd_make_external_pointer), no new bdd_nodes will be created, and hence, a garbage collection
     * will not occur.  The same goes for the array operand_nodes.
     */
    
    result = bdd_bf_and_multi_rec(mgr, operand_nodes);
    if (fml_type == BDD_FML_OR) {
	result = BDD_NOT(result);
    }
    result_bdd = bdd_make_external_pointer(mgr, result, "bdd_bf_and_or_multi");

    /*
     * Reset the cache. As part of resetting the cache, all the cache entries are freed.  A 
     * cache entry consists of an array of bdd_nodes, as well as a single bdd_node.  The 
     * array_t is freed.  In particular, the array operand_nodes will be freed by this procedure.
     */
    bdd_bf_cache_reset(mgr, FALSE /* do not reset stats */);
    bfbdd_info->sort_cache_at_gc = FALSE;

    return (result_bdd);
}


/*
 * This function is provided for convenience when building the AND of an array of BDDs.
 */
bdd_t *
bdd_and_multi(operands, build_type)
array_t *operands; /* bdd_t * */
bdd_build_type_t build_type;
{
    bdd_t *temp1_bdd, *temp2_bdd, *product_bdd;
    int i;

    if (array_n(operands) == 0) {
	fail("bdd_and_multi: empty operands array");
    }

    switch(build_type) {
    case BDD_BUILD_BINARY:
	temp1_bdd = array_fetch(bdd_t *, operands, 0);  /* get the first one */
	for (i = 1; i < array_n(operands); i++) {
	    temp2_bdd = array_fetch(bdd_t *, operands, i);  
	    product_bdd = bdd_and(temp1_bdd, temp2_bdd, 1, 1);
	    bdd_free(temp1_bdd);
	    bdd_free(temp2_bdd);
	    temp1_bdd = product_bdd;
	}
	return product_bdd;
    case BDD_BUILD_MULTI:
	/* 
	 * Could call bdd_build_formula_multi with the appropriate formula tree.
	 * However, it is much more efficient to use a routine customized
	 * for an AND array, which doesn't have to deal with cofactor_info and
	 * which has a simple termination condition.
	 */
	return bdd_bf_and_or_multi(operands, BDD_FML_AND);
	break;
    default:
	fail("bdd_and_multi_formula: unknown build type");
    }
}

/*
 * This function is provided for convenience when building the OR of an array of BDDs.
 */
bdd_t *
bdd_or_multi(operands, build_type)
array_t *operands; /* bdd_t * */
bdd_build_type_t build_type;
{
    bdd_t *temp1_bdd, *temp2_bdd, *sum_bdd;
    int i;

    if (array_n(operands) == 0) {
	fail("bdd_or_multi: empty operands array");
    }

    switch(build_type) {
    case BDD_BUILD_BINARY:
	temp1_bdd = array_fetch(bdd_t *, operands, 0);  /* get the first one */
	for (i = 1; i < array_n(operands); i++) {
	    temp2_bdd = array_fetch(bdd_t *, operands, i);  
	    sum_bdd = bdd_or(temp1_bdd, temp2_bdd, 1, 1);
	    bdd_free(temp1_bdd);
	    bdd_free(temp2_bdd);
	    temp1_bdd = sum_bdd;
	}
	return sum_bdd;
    case BDD_BUILD_MULTI:
	/* 
	 * Could call bdd_build_formula_multi with the appropriate formula tree.
	 * However, it is much more efficient to use a routine customized
	 * for an OR array, which doesn't have to deal with cofactor_info and
	 * which has a simple termination condition.
	 */
	return bdd_bf_and_or_multi(operands, BDD_FML_OR);
	break;
    default:
	fail("bdd_or_multi_formula: unknown build type");
    }
}

