#include "util.h"
#include "array.h"
#include "st.h" 

#include "bdd.h"
#include "bdd_int.h"
#include "bdd_bf.h"
#include "bdd_bf_int.h"

/*
 * Construct the BDD for formula using only standard BDD operators.
 */
static bdd_t *
bdd_build_formula_binary(formula)
bdd_formula_t *formula;
{
    bdd_fml_type_t formula_type;
    bdd_t *left_bdd, *right_bdd, *return_bdd;

    formula_type = formula->type;
    switch(formula_type) {
    case BDD_FML_OPRND:
	return_bdd = bdd_dup(formula->data1.operand);
	break;
    case BDD_FML_AND:
	left_bdd =  bdd_build_formula_binary(formula->data1.l);
	right_bdd = bdd_build_formula_binary(formula->data2.r);
        return_bdd = bdd_and(left_bdd, right_bdd, 1, 1);
	bdd_free(left_bdd);
	bdd_free(right_bdd);
	break;
    case BDD_FML_OR:
	left_bdd =  bdd_build_formula_binary(formula->data1.l);
	right_bdd = bdd_build_formula_binary(formula->data2.r);
        return_bdd = bdd_or(left_bdd, right_bdd, 1, 1);
	bdd_free(left_bdd);
	bdd_free(right_bdd);
	break;
    case BDD_FML_XOR:
	left_bdd =  bdd_build_formula_binary(formula->data1.l);
	right_bdd = bdd_build_formula_binary(formula->data2.r);
        return_bdd = bdd_xor(left_bdd, right_bdd);
	bdd_free(left_bdd);
	bdd_free(right_bdd);
	break;
    case BDD_FML_COF:
	left_bdd =  bdd_build_formula_binary(formula->data1.l); /* the thing being cofactored */
	right_bdd = bdd_build_formula_binary(formula->data2.r); /* cofactoring by */ 
        return_bdd = bdd_cofactor(left_bdd, right_bdd);
	bdd_free(left_bdd);
	bdd_free(right_bdd);
	break;
    case BDD_FML_NOT:
	left_bdd =  bdd_build_formula_binary(formula->data1.l);
        return_bdd = bdd_not(left_bdd);
	bdd_free(left_bdd);
	break;
    default:
	fail("bdd_build_formula_binary: unknown formula type");
    }
    return return_bdd;

}

/*
 * This is the core routine of the code to compute the BDD for a Boolean formula.
 */
static bdd_node *
bdd_build_formula_multi_rec(mgr, operands, cur_term_case, oprnds_cof_info, cur_active_oprnds)
bdd_manager *mgr;
array_t *operands;                    /* bdd_node * */
bdd_t *cur_term_case;
array_t *oprnds_cof_info;             /* bdd_bf_cofactor_info_t * */
var_set_t *cur_active_oprnds;
{
    bfbdd_info_t *bfbdd_info;
    bdd_safeframe frame;
    bdd_safenode then_result, else_result, result, var, oprnd;
    bdd_safenode safe_operands, safe_else_nodes;
    bdd_variableId top_id;
    int i, num_operands;
    array_t *then_nodes, *else_nodes;  /* bdd_node * */
    bdd_t *new_term_case;
    var_set_t *new_active_oprnds;      /* updated list of which operands are active */


    /*
     * Start the safeframe.
     */
    bdd_safeframe_start(mgr, frame);
    bdd_safenode_declare(mgr, then_result);
    bdd_safenode_declare(mgr, else_result);
    bdd_safenode_declare(mgr, result);
    bdd_safenode_declare(mgr, var);
    bdd_safenode_declare(mgr, oprnd);  /* just to be safe */
    bdd_bf_safenode_array(mgr, safe_operands, operands);

    bfbdd_info = (bfbdd_info_t *)mgr->undef1;
    bfbdd_info->stats.bf_ops.calls++;

    /*
     * Check if the value of operands represents a terminal case.  If it is a terminal
     * case (i.e. return is TRUE), then result.node is set, but new_term_case and new_active_oprnds 
     * are not set (and hence, new_term_case and new_active_oprnds need not be freed).  If
     * it is not a terminal case, then result.node is not set, but new_term_case and 
     * new_active_oprnds are set.
     *
     * If we do hit a terminal case, the result does not go into the cache.  Thus, we will never again
     * have the chance to free operands: thus, free it now.
     */
    if (bdd_bf_test_terminal_case(mgr, operands, oprnds_cof_info, cur_term_case, &new_term_case, 
				  cur_active_oprnds, &new_active_oprnds, &result.node)) {
	bfbdd_info->stats.bf_ops.returns.trivial++;
	bdd_safeframe_end(mgr);
	array_free(operands);
	return (result.node);
    }

    /*
     * If we have already seen the same array of operands, then we are done with this 
     * branch of the recursion.
     */
    if (bdd_bf_cache_lookup(mgr, operands, &result.node)) {
	bfbdd_info->stats.bf_ops.returns.cached++;
	bdd_safeframe_end(mgr);
	bdd_free(new_term_case);
	var_set_free(new_active_oprnds);
	array_free(operands);
	return result.node;
    }

    /*
     * Determine the minimum of the top variable IDs of the operands.  Assume BDD_ONE_ID acts like 
     * plus infinity.
     */
    top_id = (bdd_variableId) BDD_ONE_ID;
    num_operands = array_n(operands);
    for (i = 0; i < num_operands; i++) {
	oprnd.node = array_fetch(bdd_node *, operands, i);  
	top_id = MIN(top_id, BDD_REGULAR(oprnd.node)->id);
    }

    /*
     * Based on the value of top_id and the oprnds_cof_info, apply Shannon expansion to the nodes 
     * in operands.
     */
    then_nodes = array_alloc(bdd_node *, num_operands);
    else_nodes = array_alloc(bdd_node *, num_operands);
    bdd_get_branches_array(top_id, operands, then_nodes, else_nodes, oprnds_cof_info);

    /*
     * The garbage collector doesn't allow a node to be protected more than once.  Since this function
     * recursively protects its input, operands, we have to be careful how we treat the arrays then_nodes
     * and else_nodes.  It's assumed that from the point when then_nodes is set (in bdd_get_branches_array)
     * to the point it is read (in the call to bdd_build_formula_multi_rec), a garbage collection can't be
     * triggered.  However, the else_nodes array is different.  It must be protected during the first call
     * to bdd_build_formula_multi_rec, and then unprotected before its call to bdd_build_formula_multi_rec.
     * NOTE: because of this, the order of calls to bdd_build_formula_multi_rec is important.
     */
    bdd_bf_safenode_array(mgr, safe_else_nodes, else_nodes);  
    then_result.node = bdd_build_formula_multi_rec(mgr, then_nodes, new_term_case, oprnds_cof_info, new_active_oprnds);
    safe_else_nodes.array = NIL(array_t);
    else_result.node = bdd_build_formula_multi_rec(mgr, else_nodes, new_term_case, oprnds_cof_info, new_active_oprnds);

    /*
     * It is tempting to try and free the then_nodes and else_nodes arrays.  However, they have been
     * inserted into the bf_cache, so freeing them will cause bad behavior.  These arrays are 
     * eventually freed by bdd_bf_cache_reset.
     */

    /* 
     * Combine the results of the recursion down the then and else branches.  Note that this is the only
     * location where a BDD node can be created in mgr.
     */
    var.node = bdd_find_or_add(mgr, top_id, BDD_ONE(mgr), BDD_ZERO(mgr));
    result.node = bdd__ITE_(mgr, var.node, then_result.node, else_result.node);

    /*
     * Insert the result into the cache.
     */
    bdd_bf_cache_insert(mgr, operands, result.node);
    bfbdd_info->stats.bf_ops.returns.full++;

    /*
     * Clean up, and return.
     */
    bdd_safeframe_end(mgr);
    bdd_free(new_term_case);
    var_set_free(new_active_oprnds);
    return (result.node);
}

/*
 * This function sets up the recursion for computing the Boolean formula.
 */
static bdd_t *
bdd_build_formula_multi(mgr, operands, formula_tree)
bdd_manager *mgr;
array_t *operands;                 /* bdd_t * */
bdd_formula_t *formula_tree;
{
    bdd_t *terminal_case;
    int i, num_operands, num_eqn_vars;
    bdd_t *oprnd, *result_bdd;
    array_t *operand_nodes;        /* bdd_node * */
    bdd_node *result;
    array_t *oprnds_cof_info;      /* bdd_bf_cofactor_info_t * */
    bdd_bf_cofactor_info_t *cofactor_info;
    bfbdd_info_t *bfbdd_info;
    var_set_t *active_oprnds;

    /*
     * Build the BDD that represents the terminal case for the top-level formula.
     */
    terminal_case = bdd_bf_build_terminal_case_formula(formula_tree);

    /*
     * Set up the cofactor information.  This is used in get_branches to tell us which 
     * way to go.
     */
    bfbdd_info = (bfbdd_info_t *)mgr->undef1;
    bfbdd_info->cofactor_info = array_alloc(bdd_bf_cofactor_info_t *, 0);
    oprnds_cof_info = array_alloc(bdd_bf_cofactor_info_t *, 0); /* will have one cofactor_info per operand */
    cofactor_info = ALLOC(bdd_bf_cofactor_info_t, 1);           /* seed for recursive routine */
    cofactor_info->active = var_set_new(bdd_num_vars(mgr));     /* every variable initially inactive */
    cofactor_info->phase  = var_set_new(bdd_num_vars(mgr));
    bdd_bf_get_cofactor_info(mgr, formula_tree, oprnds_cof_info, cofactor_info);

    /* 
     * Convert operands into an array of bdd_nodes for recursion.
     */
    num_operands = array_n(operands);
    operand_nodes = array_alloc(bdd_node *, num_operands);
    for (i = 0; i < num_operands; i++) {
	oprnd = array_fetch(bdd_t *, operands, i);
	array_insert(bdd_node *, operand_nodes, i, oprnd->node);
    }
    
    /* 
     * Set up the var_set indicating which operands are active.  Initially, all operands are active.
     * An operands becomes inactive during the recursion when it becomes a constant, or when its
     * value no longer matters.
     */
    num_eqn_vars = array_n(bfbdd_info->eqn_vars);
    active_oprnds = var_set_new(num_eqn_vars);  /* initializes all to zero */
    /* flip the first num_operand bits to one; the rest should never be looked at, but we do this to be safe */
    for (i = 0; i < num_operands; i++) {
	var_set_set_elt(active_oprnds, i);
    }
    
    /*
     * Build the BDD for the result.  Note that we are not protecting (using safenode, etc.) the variable
     * "result".  It's assumed that from the point that result is set, to the point where it is used
     * (inside bdd_make_external_pointer), no new bdd_nodes will be created, and hence, a garbage collection
     * will not occur.  The same goes for the array operand_nodes.
     */
    result = bdd_build_formula_multi_rec(mgr, operand_nodes, terminal_case, oprnds_cof_info, active_oprnds);
    result_bdd = bdd_make_external_pointer(mgr, result, "bdd_build_formula_multi");

    /*
     * Reset the cache. As part of resetting the cache, all the cache entries are freed.  A 
     * cache entry consists of an array of bdd_nodes, as well as a single bdd_node.  The 
     * array_t is freed.  In particular, the array operand_nodes will be freed by this procedure.
     */
    bdd_bf_cache_reset(mgr, FALSE /* do not reset stats */);

    /*
     * Free up data structures.
     */
    var_set_free(active_oprnds);
    bdd_bf_clear_cofactor_info(mgr); /* clear out all cofactor_info from manager */
    bdd_free(terminal_case);
    array_free(oprnds_cof_info);

    return (result_bdd);
}

/*
 * Entry point into bdd_bf.  Build a formula tree from the formula_array and operands, and
 * then call the appropriate build routine based on build_type.
 */
bdd_t *
bdd_build_formula(formula_array, operands, build_type) 
array_t *formula_array;  /* bdd_fml_type_t */
array_t *operands;       /* bdd_t * */
bdd_build_type_t build_type;
{
    bdd_formula_t *formula_tree;
    bdd_t *temp_bdd;
    bdd_manager *mgr;

    if (array_n(operands) == 0) {
	fail("bdd_build_formula: empty operands array");
    }

    /* 
     * Get the manager via the manager field of any bdd_t.  Then check on the eqn_vars array of the manager.
     * NOTE: the eqn_vars are not needed if build_type is BINARY, so this is unnecessary overhead.
     */
    temp_bdd = array_fetch(bdd_t *, operands, 0);
    mgr = bdd_get_manager(temp_bdd);
    bdd_bf_get_equation_vars(mgr, array_n(operands));

    /*
     * Construct a tree representing the Boolean formula to be built.
     */
    formula_tree = bdd_bf_construct_formula_tree(mgr, formula_array, operands);

#if defined(BDD_DEBUG_BF) /* { */
    bdd_bf_print_formula(formula_tree);
#endif /* } */

    /*
     * Build the BDD for the formula.
     */
    switch(build_type) {
    case BDD_BUILD_BINARY:
	return (bdd_build_formula_binary(formula_tree));
    case BDD_BUILD_MULTI:
	return (bdd_build_formula_multi(mgr, operands, formula_tree));
    default:
	fail("bdd_build_formula: unknown build type");
    }

    /*
     * Free the formula tree.
     */
    bdd_bf_free_formula(formula_tree);
}


