#include "util.h"
#include "array.h"
#include "st.h" 

#include "bdd.h"
#include "bdd_int.h"
#include "bdd_bf.h"
#include "bdd_bf_int.h"

static boolean array_equal_p();
static void resize_bf_cache();
static int bdd_bf_hash();

/* 
 * The code in this file is largely borrowed from the BDD code for the ITE cache.
 */

/*
 * TODO: temporary function for experimental purposes.
 */
void
bdd_ITE_cache_reset(manager)
bdd_manager *manager;
{
    int i;
    bdd_hashcache_entry *entry;

    /*
     * Free the entries of the ITE cache, but don't free the cache itself.  
     * Nil out the buckets.
     */
    for (i = 0; i < manager->heap.cache.itetable.nbuckets; i++) {
	entry = manager->heap.cache.itetable.buckets[i];
	if (entry != NIL(bdd_hashcache_entry)) {
	    FREE(entry);
	    manager->heap.cache.itetable.buckets[i] = NIL(bdd_hashcache_entry);
	}
    }
    manager->heap.cache.itetable.nentries = 0;
    return;
}

/*
 * Clear out the contents of the bf_cache.  It's assumed that the bf_cache is reset
 * after each top-level bdd_build_formula operation.
 */
void
bdd_bf_cache_reset(manager, clear_stats)
bdd_manager *manager;
boolean clear_stats;
{
    bfbdd_info_t *bfbdd_info;
    int i;
    bdd_bf_cache_entry_t *bf_entry;

    bfbdd_info = (bfbdd_info_t *)manager->undef1;

    /*
     * Free the entries of the bf_cache, but don't free the cache itself.  
     * Nil out the buckets.
     */
    for (i = 0; i < bfbdd_info->bf_cache.nbuckets; i++) {
	bf_entry = bfbdd_info->bf_cache.buckets[i];
	if (bf_entry != NIL(bdd_bf_cache_entry_t)) {
	    array_free(bf_entry->operands);
	    FREE(bf_entry);
	    bfbdd_info->bf_cache.buckets[i] = NIL(bdd_bf_cache_entry_t);
	}
    }
    bfbdd_info->bf_cache.nentries = 0;

    if (clear_stats == TRUE) {
	/*
	 *    Clear out the stats
	 */
	bfbdd_info->stats.bf_cache.hits = 0;
	bfbdd_info->stats.bf_cache.misses = 0;
	bfbdd_info->stats.bf_cache.collisions = 0;
	bfbdd_info->stats.bf_cache.inserts = 0;

	bfbdd_info->stats.bf_ops.calls = 0;
	bfbdd_info->stats.bf_ops.returns.trivial = 0;
	bfbdd_info->stats.bf_ops.returns.cached = 0;
	bfbdd_info->stats.bf_ops.returns.full = 0;

	bfbdd_info->stats.memory.bf_cache = 0;
    }

    return;
}

/*
 * Insert into the bf_cache.  The key of an entry is an array of bdd_nodes, and
 * the value is a single bdd_node.
 */
void
bdd_bf_cache_insert(manager, operands, data)
bdd_manager *manager;
array_t *operands;   /* bdd_node * */
bdd_node *data;
{
    bfbdd_info_t *bfbdd_info;
    int pos;
    bdd_bf_cache_entry_t *entry;

    /* 
     * If the bf_cache is turned off, just return.
     */
    bfbdd_info = (bfbdd_info_t *)manager->undef1;
    if (bfbdd_info->bf_cache.on == FALSE) {
	return;		
    }

    /*
     * Calculate which bucket this entry is to go into, and get the current contents of this bucket.
     */
    pos = bdd_bf_hash(manager, operands);
    entry = bfbdd_info->bf_cache.buckets[pos];

    if (entry != NIL(bdd_bf_cache_entry_t)) {
	/*
	 * An entry already exists at pos.  Since we are using an open-addressing
	 * scheme, we will just reuse the memory of this entry.  Number of entries
	 * stays constant.  Must free the array of operands in the entry.
	 */
	array_free(entry->operands);
	bfbdd_info->stats.bf_cache.collisions++;
    } else {
	/*
	 * Must allocate a new entry, since one does not already exist at pos. It's possible
	 * that allocating a bdd_bf_cache_entry_t will cause the manager's 
         * memory limit to be exceeded. However, we allow this to happen, because the alternative
	 * is that we don't cache this call.  Not caching new calls could be deadly for runtime.
	 */
	entry = ALLOC(bdd_bf_cache_entry_t, 1);
        if (entry == NIL(bdd_bf_cache_entry_t)) {  
	    (void) bdd_memfail(manager, "bdd_bf_cache_insert");
        }
        bfbdd_info->bf_cache.buckets[pos] = entry;
	bfbdd_info->bf_cache.nentries++;
    }

    /*
     * Fill in the contents of the entry.
     */
    entry->operands = operands;
    entry->data = data;
    bfbdd_info->stats.bf_cache.inserts++;

    /*
     * Check to see if table needs resizing.  We base the decision on the ratio of the 
     * number of entries to the number of buckets.
     */
    if ( (unsigned int) (((float) bfbdd_info->bf_cache.nentries
                          / (float) bfbdd_info->bf_cache.nbuckets) * 100) 
                         > bfbdd_info->bf_cache.resize_at) {
        (void) resize_bf_cache(manager);
    }
}

/*
 *    bdd_bf_cache_lookup - perform a lookup in the cache
 *
 *    return {TRUE, FALSE} on {found, not found}.
 */
boolean
bdd_bf_cache_lookup(manager, operands, value)
bdd_manager *manager;
array_t *operands;              /* bdd_node * */
bdd_node **value;		/* return */
{
    bfbdd_info_t *bfbdd_info;
    int pos;
    bdd_bf_cache_entry_t *entry;
    array_t *entry_oprnds;      /* bdd_node * */

    /* 
     * Always fails if caching is off.
     */
    bfbdd_info = (bfbdd_info_t *)manager->undef1;
    if (bfbdd_info->bf_cache.on == FALSE) {
	return (FALSE);		
    }

    /*
     * Calculate which bucket this entry is to go into, and get the current contents of this bucket.
     */
    pos = bdd_bf_hash(manager, operands);
    entry = bfbdd_info->bf_cache.buckets[pos];

    if (entry != NIL(bdd_bf_cache_entry_t)) {
	/*
	 * Entry exists at pos.  See if it matches.  If not (a miss) return.  If it does match
	 * (a hit), then will drop out of IF and set value. Note that it's not enough for the
	 * addresses of the arrays to agree.
	 */
	entry_oprnds = entry->operands;
	assert(entry->data != NIL(bdd_node));
 	if (array_equal_p(operands, entry_oprnds) == FALSE) {  
	    bfbdd_info->stats.bf_cache.misses++;
	    return (FALSE);
	}
    } else {
	/*
	 * No entry exists at pos.  Definitely a miss.
	 */
	bfbdd_info->stats.bf_cache.misses++;
	return (FALSE);
    }  

    bfbdd_info->stats.bf_cache.hits++;

    *value = entry->data;
    return (TRUE);
}

/*
 * Make the bf_cache larger.  Perform a rehash to map all of the old
 * stuff into this new larger world. Note that since we don't maintain collision chains,
 * some stuff may be lost.
 */
static void
resize_bf_cache(manager)
bdd_manager *manager;
{
    bfbdd_info_t *bfbdd_info;
    unsigned int old_nbuckets, next_hash_prime;
    bdd_bf_cache_entry_t **old_buckets, *old_entry;
    int i, pos;

    bfbdd_info = (bfbdd_info_t *)manager->undef1;

    /*
     * If the max_size has already been exceeded, then just return.
     */
    if (bfbdd_info->bf_cache.nbuckets > bfbdd_info->bf_cache.max_size) {
        return;
    }

    /*
     * Get the next hash prime.  If allocating a new table will cause the manager memory limit to 
     * be exceeded, then just return.  Trading off time in favor of space.
     */
    next_hash_prime = bdd_get_next_hash_prime(bfbdd_info->bf_cache.nbuckets);
    if (bdd_will_exceed_mem_limit(manager, (next_hash_prime * sizeof(bdd_bf_cache_entry_t *)), FALSE) == TRUE) { 
        return;
    }

    /*
     * Save the old buckets.
     */
    old_nbuckets = bfbdd_info->bf_cache.nbuckets;
    old_buckets = bfbdd_info->bf_cache.buckets;

    /*
     * Get the new size and allocate the new table.  We are growing by roughly a factor of two.
     * If there isn't sufficient memory, don't fail: the cache
     * is meant to enhance performance, but we can still continue even if we can't resize.
     */
    bfbdd_info->bf_cache.nbuckets = next_hash_prime;
    bfbdd_info->bf_cache.buckets = ALLOC(bdd_bf_cache_entry_t *, bfbdd_info->bf_cache.nbuckets);
    if (bfbdd_info->bf_cache.buckets == NIL(bdd_bf_cache_entry_t *)) {
        bfbdd_info->bf_cache.nbuckets = old_nbuckets;
        bfbdd_info->bf_cache.buckets = old_buckets;
        return;
    }

    /* 
     * Initialize all the entries to NIL.
     */
    for (i = 0; i < bfbdd_info->bf_cache.nbuckets; i++) {
	bfbdd_info->bf_cache.buckets[i] = NIL(bdd_bf_cache_entry_t);
    }

    /*
     * Remap all of the ITE cache entries to the new larger cache.
     */
    bfbdd_info->bf_cache.nentries = 0;
    for (i = 0; i < old_nbuckets; i++) {
	old_entry = old_buckets[i];
	if (old_entry != NIL(bdd_bf_cache_entry_t)) {

	    /*
	     * Get the hash position for the entry in the new cache.  If the position is already
	     * occupied, then old_entry will not be rehashed; it's lost;  If the position is not
	     * occupied, then simply set the new cache bucket pointer to old_entry.
	     */
	    pos = bdd_bf_hash(manager, old_entry->operands);
	    if (bfbdd_info->bf_cache.buckets[pos] != NIL(bdd_bf_cache_entry_t)) {
		array_free(old_entry->operands);
		FREE(old_entry);
	    } else {
		bfbdd_info->bf_cache.buckets[pos] = old_entry;
		bfbdd_info->bf_cache.nentries++;
	    }
	}
    }

    /*
     * Free the array of buckets.  Note that the entries of the table were simply 
     * moved over to the new table.
     */
    FREE(old_buckets);
}

/*
 * Compare two arrays of bdd_nodes.  They are equal iff they are equal at every position.
 */
static boolean
array_equal_p(arr1, arr2) 
array_t *arr1;    /* assumed to be of bdd_node * */
array_t *arr2;    /* assumed to be of bdd_node * */
{
    bdd_node *node1, *node2;
    int i;

    assert(array_n(arr1) == array_n(arr2));

    for (i = 0; i < array_n(arr1); i++) {
	node1 = array_fetch(bdd_node *, arr1, i);
	node2 = array_fetch(bdd_node *, arr2, i);
	if (node1 != node2) {
	    return FALSE;
	}
    }

    return TRUE;
}

#define shift_array_length 5
static int shift_array[shift_array_length] = {5, 7, 9, 11, 13};

/*
 * Compute the hash value for an array of bdd_nodes. The hash function is simple: 
 * just add all the bdd_node addresses together, and divide by the modulus.
 */
static int
bdd_bf_hash(manager, operands)
bdd_manager *manager;
array_t *operands;    /* bdd_node * */
{
    bfbdd_info_t *bfbdd_info;
    register unsigned int sum = 0;
    register int i, j;

    bfbdd_info = (bfbdd_info_t *)manager->undef1;
    
    j = 0;
    for (i = 0; i < array_n(operands); i++) {
	sum += ((unsigned int) array_fetch(bdd_node *, operands, i)) << shift_array[j];
	if (j == (shift_array_length - 1)) {
	    j = 0;
	} else {
	    j++;
	}
    }

    return (sum % ((unsigned int) bfbdd_info->bf_cache.nbuckets));
}

/*
 * Perform garbage collection on the bf_cache.  This function is called from bdd_garbage_collect.
 */
void
bdd_bf_scan_cache(manager)
bdd_manager *manager;
{
    bfbdd_info_t *bfbdd_info;
    int i, pos;
    bdd_bf_cache_entry_t *bf_entry, **old_buckets;

    bfbdd_info = (bfbdd_info_t *)manager->undef1;

    /*
     * If bfbdd_info is NIL, it means that the bf_cache has not yet been set up,
     * so there is nothing to do.
     */
    if (bfbdd_info == NIL(bfbdd_info_t)) {
	return;
    }

    /*
     * If there are no entries in the cache, then there is nothing to scan.
     */
    if (bfbdd_info->bf_cache.nentries == 0) {
        return;
    }

    /*
     * If the cache is to be invalidated, then just clear all the entries.
     */
    if (bfbdd_info->bf_cache.invalidate_on_gc) {
	bdd_bf_cache_reset(manager, FALSE);
	return;
    }

    /*
     * Otherwise the cache must be updated through a rehashing
     * because all of the nodes in the keys in the bf_cache have been changed.
     */

    /*
     * Save the old one.
     */
    old_buckets = bfbdd_info->bf_cache.buckets;

    /*
     * Make a new clean one so we can rehash into it.
     */
    bfbdd_info->bf_cache.buckets = ALLOC(bdd_bf_cache_entry_t *, bfbdd_info->bf_cache.nbuckets);
    if (bfbdd_info->bf_cache.buckets == NIL(bdd_bf_cache_entry_t *)) {
	(void) bdd_memfail(manager, "bdd_bf_scan_cache");
    }
    for (i = 0; i < bfbdd_info->bf_cache.nbuckets; i++) {
	bfbdd_info->bf_cache.buckets[i] = NIL(bdd_bf_cache_entry_t);
    }

    /*
     * Rehash the entries.
     */
    bfbdd_info->bf_cache.nentries = 0;
    for (i = 0; i < bfbdd_info->bf_cache.nbuckets; i++) {
	bf_entry = old_buckets[i];	/* to make it textually easier to read */
	if (bf_entry != NIL(bdd_bf_cache_entry_t)) {

	    /*
	     * The update policy is easy.  Since bfbdd guarantees that no unnecessary bdd_nodes
	     * are created, then all the nodes referenced by the cache are alive.  Note that 
	     * even if this assumption is violated, it just means that dead nodes will be 
	     * treated as live nodes (i.e. entries with live nodes will be saved).
	     */
	    bdd_bf_relocate_entry_contents(manager, bf_entry); 

	    /*
	     * If we are keeping the arrays sorted (by the addresses of the bdd_nodes in the arrays),
	     * then we must resort, since relocating upsets the sort.
	     */
	    if (bfbdd_info->sort_cache_at_gc == TRUE) {
		array_sort(bf_entry->operands, bdd_bf_bdd_node_compare); 
	    }

	    /*
	     * Get the hash position for the entry in the new cache.  If the position is already
	     * occupied, then bf_entry will not be rehashed; it's lost;  If the position is not
	     * occupied, then we can simply reuse the memory of the old entry.  We cannot use
	     * bdd_hashcache_insert because it doesn't work when the cache is turned off.
	     */
	    pos = bdd_bf_hash(manager, bf_entry->operands);
	    if (bfbdd_info->bf_cache.buckets[pos] != NIL(bdd_bf_cache_entry_t)) {
		array_free(bf_entry->operands);
		FREE(bf_entry);
	    } else {
		bfbdd_info->bf_cache.buckets[pos] = bf_entry;
		bfbdd_info->bf_cache.nentries++;
	    }

	} /* close: if (bf_entry != NIL(bdd_bf_cache_entry_t)) { */
    } /* close: for (i = 0; i < bfbdd_info->bf_cache.nbuckets; i++) { */

    /*
     * We are done rehashing. Get rid of the old_buckets.
     */
    FREE(old_buckets);
}

