#include "util.h"
#include "array.h"
#include "st.h" 

#include "bdd.h"
#include "bdd_int.h"
#include "bdd_bf.h"
#include "bdd_bf_int.h"

/* 
 * Get the then and else branches of the bdd_nodes in the array operands (modulo the top_id
 * and cofactor info).
 */
void
bdd_get_branches_array(top_id, operands, then_nodes, else_nodes, oprnds_cof_info)
bdd_variableId top_id;
array_t *operands;	   /* bdd_node * */	
array_t *then_nodes;       /* bdd_node *, return */
array_t *else_nodes;       /* bdd_node *, return */
array_t *oprnds_cof_info;  /* bdd_bf_cofactor_info_t * */
{
    int i;
    bdd_node *oprnd, *oprnd_0, *oprnd_1;
    bdd_variableId oprnd_id;
    bdd_bf_cofactor_info_t *cofactor_info;

    /*
     * NOTE: nodes in this function are not protected for garbage collection.  We assume that 
     * no operations are invoked from this function which could cause a bdd_node to be created.
     */

    for (i = 0; i < array_n(operands); i++) {
	oprnd = array_fetch(bdd_node *, operands, i);
	oprnd_id = BDD_REGULAR(oprnd)->id;

	if (top_id < oprnd_id) {
	    /*
	     * oprnd is "below" top_id, so we don't perform Shannon expansion.
	     */
	    array_insert(bdd_node *, then_nodes, i, oprnd);
	    array_insert(bdd_node *, else_nodes, i, oprnd);
	} else if (top_id == oprnd_id) {
	    /* 
	     * oprnd_id and top_id correspond.  First get the branches of oprnd.  Then,
	     * get the cofactor_info for this oprnd, and determine if this oprnd is being 
	     * cofactored by top_id.
	     */
	    bdd_get_branches(oprnd, &oprnd_1, &oprnd_0);
	    if (oprnds_cof_info != NIL(array_t)) {
		cofactor_info = array_fetch(bdd_bf_cofactor_info_t *, oprnds_cof_info, i); 
		if (var_set_get_elt(cofactor_info->active, top_id)) {
		    /* need to cofactor by top_id */
		    if (var_set_get_elt(cofactor_info->phase, top_id)) {
			/* positive cofactor */
			array_insert(bdd_node *, then_nodes, i, oprnd_1);
			array_insert(bdd_node *, else_nodes, i, oprnd_1);
		    } else {
			/* negative cofactor */
			array_insert(bdd_node *, then_nodes, i, oprnd_0);
			array_insert(bdd_node *, else_nodes, i, oprnd_0);
		    }
		} else {
		    /* split on top_id */
		    array_insert(bdd_node *, then_nodes, i, oprnd_1);
		    array_insert(bdd_node *, else_nodes, i, oprnd_0);
		}
	    } else {
		/* split on top_id */
		array_insert(bdd_node *, then_nodes, i, oprnd_1);
		array_insert(bdd_node *, else_nodes, i, oprnd_0);
	    }
	} else {
	    fail("bdd_get_branches_array: top_id > oprnd_id, which is impossible");
	}
    }
}

/*
 * Each distinct cofactor_info_t structure is stored with the manager.  This function
 * determines if there already exists a cofactor_info_t in the manager which is 
 * equivalent to the input cofactor_info.  If not it inserts in_cofactor_info into the manager.
 * After the entire formula is processed, the number of distinct cofactor_info_t objects is 
 * bounded by the number of COF ops in the formula.
 *
 * The reason for storing just one copy of each distinct cofactor_info, rather than letting
 * each operand have its own copy, is that I suspect that there will be many operands with
 * identical cofactor_infos (consider the special case where there are no COF ops - then
 * all the cofactor_infos will be the same), and there may be many operands.
 */
static void
insert_cofactor_info_in_mgr(mgr, in_cofactor_info)
bdd_manager *mgr;
bdd_bf_cofactor_info_t *in_cofactor_info;
{
    bfbdd_info_t *bfbdd_info;
    int i;
    bdd_bf_cofactor_info_t *mgr_cofactor_info;

    bfbdd_info = (bfbdd_info_t *)mgr->undef1;

    /*
     * Check each cofactor_info stored in the manager to see if in_cofactor_info already exists.
     */
    for (i = 0; i < array_n(bfbdd_info->cofactor_info); i++) {
	mgr_cofactor_info = array_fetch(bdd_bf_cofactor_info_t *, bfbdd_info->cofactor_info, i);
	if (var_set_equal(mgr_cofactor_info->active, in_cofactor_info->active)) {
	    if (var_set_equal(mgr_cofactor_info->phase, in_cofactor_info->phase)) {
		/* 
		 * Found a match.  By construction, in_cofactor_info and mgr_cofactor_info
		 * should be the same structure.
		 */
		assert(mgr_cofactor_info == in_cofactor_info);
		return;
	    }
	}
    }

    /*
     * Didn't find a match.  Add in_cofactor_info and return.
     */
    array_insert_last(bdd_bf_cofactor_info_t *, bfbdd_info->cofactor_info, in_cofactor_info); 
    return;
}

/*
 * Recursive routine to set those bits in cofactor_info corresponding to the variables appearing 
 * in f.  It's assumed that f is a cube.  It's possible that a variable appearing in f has already
 * been set in cofactor_info.  This is fine, as long as the variable appears in the same phase.
 */
static void
add_to_cofactor_info_rec(mgr, cofactor_info, f)
bdd_manager *mgr;
bdd_bf_cofactor_info_t *cofactor_info;
bdd_node *f;
{
    bdd_node *f0, *f1;
    boolean already_active = FALSE;
    int phase;
    bdd_variableId f_id;

    /* 
     * NOTE: The bdd_nodes in this routine are not protected against a garbage collection.  It is
     * assumed that no bdd_nodes are created during any routine is the transitive closure of this one.
     */

    BDD_ASSERT(f != BDD_ZERO(mgr));

    /*
     * Terminal condition of recursion: no more variables left in f.
     */
    if (f == BDD_ONE(mgr)) {
	return;
    }

    /*
     * Set the bit in the active var_set corresponding to the id of f.  Take note if the bit
     * was already set.
     */
    f_id = BDD_REGULAR(f)->id;
    if (var_set_get_elt(cofactor_info->active, (int) f_id)) {
	already_active = TRUE;
	/* note that phase is later read only if already_active is TRUE */
	phase = var_set_get_elt(cofactor_info->phase, (int) f_id);
    } else {
	var_set_set_elt(cofactor_info->active, (int) f_id);
    }

    /*
     * Get the then and else branches of f.  Exactly one of the branches must point to ZERO to
     * satisfy the condition that f is a cube.
     */
    bdd_get_branches(f, &f1, &f0);
    if (f1 == BDD_ZERO(mgr)) {
	/* cofactor on negative phase */
	if (already_active == TRUE) {
	    if (phase == 1) {
		fail("add_to_cofactor_info_rec: trying to cofactor on both phases of a variable");
            }		
	} /* else: f_id is present in negative phase; keep phase bit at zero */
	add_to_cofactor_info_rec(mgr, cofactor_info, f0);
    } else if (f0 == BDD_ZERO(mgr)) {
	/* cofactor on positive phase */
	if (already_active == TRUE) {
	    if (phase == 0) {
		fail("add_to_cofactor_info_rec: trying to cofactor on both phases of a variable");
            } 
	} else {
	    var_set_set_elt(cofactor_info->phase, (int) f_id);  /* f_id is present in positive phase */
	}
	add_to_cofactor_info_rec(mgr, cofactor_info, f1);
    } else { /* not a cube, because neither branch is zero */
	fail("add_to_cofactor_info_rec: cofactor operand is not a cube");
    }
}

/*
 * Set those bits in cofactor_info corresponding to the variables appearing in the 
 * formula.  It's assumed that formula is of type OPRND, and furthermore, that the
 * operand is a cube.
 */
static void
add_to_cofactor_info(mgr, cofactor_info, formula)
bdd_manager *mgr;
bdd_bf_cofactor_info_t *cofactor_info;
bdd_formula_t *formula;
{
    bdd_t *oprnd;

    if (formula->type != BDD_FML_OPRND) {
	fail("add_to_cofactor_info: cofactor operand must be a BDD; cannot be an expression");
    }

    /*
     * Must traverse the BDD, determining which variables are present, and in which phase.
     * Fail if BDD doesn't represent a cube.
     */
    oprnd = formula->data1.operand;
    add_to_cofactor_info_rec(mgr, cofactor_info, oprnd->node);
}

/*
 * Cofactor operations in a formula affect what branches to take in the BDDs during recursion.
 * We want to compute this information once and store it in a efficient manner.  oprnds_cof_info
 * has a pointer to a bdd_bf_cofactor_info_t structure for each operand.  A bdd_bf_cofactor_info_t 
 * structure consists of two var_sets, active and phase, each of length the number of variables in 
 * the bdd_manager.  Thus, the jth bit of the active var_set, of the ith operand, tells us whether
 * or not the ith operand is to be cofactored by the jth variable.  If so, then the jth bit of 
 * the phase var_set tells us in which phase the cofactor is.  
 *
 * This function is initially called with an empty array for oprnds_cof_info, and var_sets with
 * no bits set for current_cof_info (this means that at the top level, we haven't seen any 
 * cofactors yet).  As we reach OPRNDs while recursing on formula, we fill up the oprnds_cof_info.
 * At each point in the recursion, the current_cof_info tells us, for each operand, what variables
 * are to be cofactored so far.
 *
 * TODO: could merge this function with build_terminal_case_formula, so that 'formula' is traversed
 * only once.
 */
void
bdd_bf_get_cofactor_info(mgr, formula, oprnds_cof_info, current_cof_info)
bdd_manager *mgr;
bdd_formula_t *formula;
array_t *oprnds_cof_info;		/* of bdd_bf_cofactor_info_t * */
bdd_bf_cofactor_info_t *current_cof_info;
{
    bdd_fml_type_t formula_type;
    bdd_bf_cofactor_info_t *new_cof_info;

    formula_type = formula->type;
    switch(formula_type) {
    case BDD_FML_OPRND:
	/*
	 * We have reached the terminal condition of the recursion.  We thus have all the 
	 * cofactor information for this operand.  We want to store just one copy of each 
	 * unique cofactor_info; the following function keeps track of all the cofactor_info's
	 * created thus far.
	 */
	insert_cofactor_info_in_mgr(mgr, current_cof_info);

	/*
	 * It's critical that the ith location in the array oprnds_cof_info corresponds to 
	 * the ith location in the original operands array.  This is guarenteed by traversing
	 * the formula in the same manner in which is was created.
	 */
	array_insert_last(bdd_bf_cofactor_info_t *, oprnds_cof_info, current_cof_info);  
	break;
    case BDD_FML_AND:
    case BDD_FML_OR:
    case BDD_FML_XOR:
	/*
	 * Since these operators don't involve cofactoring, we just pass along current_cof_info
	 * to the sub-operands.
	 */
	bdd_bf_get_cofactor_info(mgr, formula->data1.l, oprnds_cof_info, current_cof_info);
	bdd_bf_get_cofactor_info(mgr, formula->data2.r, oprnds_cof_info, current_cof_info);
	break;
    case BDD_FML_COF:
	/*
	 * We have hit a cofactor operation.  Since we will be setting additional bits in the 
	 * current_cof_info, we need to first make a copy of current_cof_info (we can't modify
	 * current_cof_info directly, because it is being used on other branches of the recursion).
	 */
	new_cof_info = ALLOC(bdd_bf_cofactor_info_t, 1);
	new_cof_info->active = var_set_copy(current_cof_info->active);
	new_cof_info->phase  = var_set_copy(current_cof_info->phase);

	/*
	 * Set those bits in new_cof_info corresponding to the variables appearing in the 
	 * cube by which we are cofactoring.  Remember, the term we are cofactoring by is 
	 * the right operand.  Note that we allow an operand to be cofactored more than once 
	 * by the same variable in the same phase, but not in opposite phases.
	 */
	add_to_cofactor_info(mgr, new_cof_info, formula->data2.r); 

	/*
	 * Continue the recursion with the new_cof_info.
	 */
	bdd_bf_get_cofactor_info(mgr, formula->data1.l, oprnds_cof_info, new_cof_info);
	bdd_bf_get_cofactor_info(mgr, formula->data2.r, oprnds_cof_info, new_cof_info);
	break;
    case BDD_FML_NOT:
	bdd_bf_get_cofactor_info(mgr, formula->data1.l, oprnds_cof_info, current_cof_info);
	break;
    default:
	fail("bdd_bf_get_cofactor_info: unknown formula type");
    }
    return;
}

/*
 * The cofactor_info is specific to a given formula.  Thus, when we are done processing a 
 * formula, we must clear all the current cofactor_info from the manager.
 */
void
bdd_bf_clear_cofactor_info(mgr)
bdd_manager *mgr;
{
    bfbdd_info_t *bfbdd_info;
    bdd_bf_cofactor_info_t *cofactor_info;
    int i;

    bfbdd_info = (bfbdd_info_t *)mgr->undef1;

    for (i = 0; i < array_n(bfbdd_info->cofactor_info); i++) {
	cofactor_info = array_fetch(bdd_bf_cofactor_info_t *, bfbdd_info->cofactor_info, i);
	var_set_free(cofactor_info->active);
	var_set_free(cofactor_info->phase);
	FREE(cofactor_info);
    }
}


