/*
 * Internal bdd_bf data structures.
 */

/* 
 * A node in the formula tree. Used to represent a Boolean expression.
 */
typedef struct bdd_formula {
    bdd_fml_type_t type;        /* type of this node */
    union {
	struct bdd_formula *l;	/* left child in Boolean formula */
	bdd_t *operand;         /* the BDD, if type is OPRND */
    } data1;
    union {
	struct bdd_formula *r;	/* right child, or null if not used */
	bdd_t *eqn_var;         /* the eqn_var, if type is OPRND */
    } data2;
} bdd_formula_t;

/*
 * Stats about bf_cache and bf_ops.  Meant to have same structure as bdd_stats.
 */
typedef struct bdd_bf_stats {
    bdd_cache_stats bf_cache;
    struct {
	unsigned int calls;
	struct {
	    unsigned int trivial;
	    unsigned int cached;
	    unsigned int full;
	} returns;
    } bf_ops;
    struct {
	unsigned int bf_cache;
    } memory;           /* memory usage */
} bdd_bf_stats_t;

/*
 * One cofactor_info is maintained for each operand, so that we know during
 * recursion what the cofactor variables are for the operand.
 */
typedef struct bdd_bf_cofactor_info {
    var_set_t *active;                      /* is the ith variable involved in a cofactor */      
    var_set_t *phase;                       /* if the ith variable is active, what is its phase */
} bdd_bf_cofactor_info_t;

/*
 * Entry for the bf_cache; used to remember previous computations.
 */
typedef struct bdd_bf_cache_entry {
    array_t *operands;                      /* array of bdd_node * */
    bdd_node *data;		
} bdd_bf_cache_entry_t;

/*
 * Various global variables needed for the bdd_bf code.  NOTE: may want to eventually
 * put these directly into the bdd_manager.
 */
typedef struct bfbdd_info {
    array_t *eqn_vars;                      /* the equation vars, bdd_t * */
    bdd_manager *ev_mgr;                    /* the manager for the eqn_vars */
    array_t *cofactor_info;                 /* array of distinct bdd_bf_cofactor_info_t * */
    boolean sort_cache_at_gc;               /* flag to indicate if arrays in bf_cache should be sorted after gc; dflt is FALSE */
    struct {
	boolean on;			    /* is caching even turned on? */
	boolean invalidate_on_gc;	    /* always invalidate the cache on a gc */
	unsigned int resize_at;             /* percentage at which to resize (e.g. 85% is 85) */
	unsigned int max_size;              /* don't resize if nbuckets exceeds this */
	unsigned int nbuckets;	            /* buckets in the cache */
	unsigned int nentries;	            /* entries in the cache */
	bdd_bf_cache_entry_t **buckets;     /* the cache */
    } bf_cache;		                    /* cache of the Boolean formula entries */
    bdd_bf_stats_t stats;
} bfbdd_info_t;

/*
 * Internal functions.
 */
EXTERN bdd_formula_t *bdd_bf_construct_formula_tree ARGS((bdd_manager *, array_t *, array_t *));
EXTERN void bdd_bf_print_formula ARGS((bdd_formula_t *));
EXTERN void bdd_bf_free_formula ARGS((bdd_formula_t *));
EXTERN void bdd_bf_cache_reset ARGS((bdd_manager *, boolean));
EXTERN void bdd_bf_cache_insert ARGS((bdd_manager *, array_t *, bdd_node *));
EXTERN boolean bdd_bf_cache_lookup ARGS((bdd_manager *, array_t *, bdd_node **));
EXTERN bdd_t *bdd_bf_build_terminal_case_formula ARGS((bdd_formula_t *));
EXTERN void bdd_bf_get_equation_vars ARGS((bdd_manager *, int));
EXTERN void bdd_bf_get_cofactor_info ARGS((bdd_manager *, bdd_formula_t *, array_t *, bdd_bf_cofactor_info_t *));
EXTERN void bdd_bf_clear_cofactor_info ARGS((bdd_manager *));
EXTERN void bdd_bf_scan_cache ARGS((bdd_manager *));
EXTERN void bdd_bf_relocate_entry_contents ARGS((bdd_manager *, bdd_bf_cache_entry_t *));
EXTERN void bdd_bf_relocate_array ARGS((bdd_manager *, array_t *, boolean));
EXTERN void bdd_get_branches_array ARGS((bdd_variableId, array_t *, array_t *, array_t *, array_t *));
EXTERN bdd_t *bdd_bf_and_or_multi ARGS((array_t *, bdd_fml_type_t));
EXTERN int bdd_bf_bdd_node_compare ARGS((char *, char *));
EXTERN boolean bdd_bf_test_terminal_case ARGS((bdd_manager *, array_t *, array_t *, bdd_t *, bdd_t **, 
					       var_set_t *, var_set_t **, bdd_node **));

/*
 * Extra macro to handle arrays of bdd_nodes.
 */
#define bdd_bf_safenode__array(manager, sn, arr)  \
                    { (sn).node = NIL(bdd_node);  \
		      (sn).arg = NIL(bdd_node *); \
		      (sn).array = arr;           \
		      (sn).next = (manager)->heap.internal_refs.frames->nodes; \
		      (manager)->heap.internal_refs.frames->nodes = &(sn); }

/*
 *    void
 *    bdd_bf_safenode_array(manager, sn, arr)
 *    bdd_manager *manager;
 *    bdd_safenode sn;		
 *    array_t *arr;		
 *
 */
#define bdd_bf_safenode_array(manager, sn, arr)	\
		    { bdd_bf_safenode__array(manager, sn, arr); \
		      BDD_ASSERT_FRAMES_CORRECT(manager); }
