#include "util.h"
#include "array.h"
#include "st.h" 

#include "bdd.h"
#include "bdd_int.h"
#include "bdd_bf.h"
#include "bdd_bf_int.h"


/*
 * Make sure there are enough equation variables in the manager.  For each operand
 * in the original formula, we associate a single, binary-valued variable.  These
 * variable are reused from formula to formula (the ith eqn_var always corresponds
 * to the ith operand).  We create these variables in a separate manager for two
 * reasons: 1) so that we can be assured that the ith eqn_var has variableId i, and
 * 2) so that we can easily account for the overhead (time and space) associated
 * with building and testing the terminal condition.
 */
void
bdd_bf_get_equation_vars(mgr, num_needed)
bdd_manager *mgr;
int num_needed;
{
    bfbdd_info_t *bfbdd_info;
    int i, num_vars;
    array_t *eqn_vars_array;    /* bdd_t * */
    bdd_t *eqn_var;

    bfbdd_info = (bfbdd_info_t *) mgr->undef1;
    eqn_vars_array = bfbdd_info->eqn_vars;
    num_vars = array_n(eqn_vars_array);

    if (num_vars < num_needed) {
	/* 
	 * Must make up the difference.
	 */
	for (i = num_vars; i < num_needed; i++) {
	    /* create the eqn_vars in a different manager */
	    eqn_var = bdd_create_variable(bfbdd_info->ev_mgr);  
	    array_insert_last(bdd_t *, eqn_vars_array, eqn_var);
	}
    }
}

/*
 * Build the BDD which determines if a terminal case has been reached
 * in building the BDD for a formula.  Conceptually, this is done by
 * replacing each operand of the original formula by a single, 
 * binary-valued variable, call the eqn_var.  Then we just build the
 * BDD for this Boolean expression.  The result is the "characteristic
 * function" for the expression.
 *
 * NOTE: this code is identical to bdd_build_formula_binary.  The only 
 * differences are: 1) OPRND case uses eqn_var, 2) COF case (actually, 
 * there would be no effect in performing the cofactor).
 */
bdd_t *
bdd_bf_build_terminal_case_formula(formula)
bdd_formula_t *formula;
{
    bdd_fml_type_t formula_type;
    bdd_t *result, *left, *right;

    formula_type = formula->type;

    /*
     * Compute the new result, based on the values from recursion.
     */
    switch(formula_type) {
    case BDD_FML_OPRND:
	result = bdd_dup(formula->data2.eqn_var);
	break;
    case BDD_FML_AND:
	left = bdd_bf_build_terminal_case_formula(formula->data1.l);
	right = bdd_bf_build_terminal_case_formula(formula->data2.r);
	result = bdd_and(left, right, 1, 1);
	bdd_free(left);
	bdd_free(right);
	break;
    case BDD_FML_OR:
	left = bdd_bf_build_terminal_case_formula(formula->data1.l);
	right = bdd_bf_build_terminal_case_formula(formula->data2.r);
	result = bdd_or(left, right, 1, 1);
	bdd_free(left);
	bdd_free(right);
	break;
    case BDD_FML_XOR:
	left = bdd_bf_build_terminal_case_formula(formula->data1.l);
	right = bdd_bf_build_terminal_case_formula(formula->data2.r);
	result = bdd_xor(left, right);
	bdd_free(left);
	bdd_free(right);
	break;
    case BDD_FML_COF:
	/*
	 * In this case, we just return the result of the recursion
	 * down the left branch.  This is because cofactoring has no
	 * impact on the characteristic function of the expression; the
	 * only place it has an impact is on chooses which branches
	 * of the BDD to recurse down.
	 */
	left = bdd_bf_build_terminal_case_formula(formula->data1.l);
	result = bdd_dup(left);
	bdd_free(left);
	break;
    case BDD_FML_NOT:
	left = bdd_bf_build_terminal_case_formula(formula->data1.l);
	result = bdd_not(left);
	bdd_free(left);
	break;
    default:
	fail("bdd_bf_build_terminal_case_formula: unknown formula type");
    }

    return result;

}

/*
 * This function determines whether or not we have reached a terminal condition in the
 * recursive routine to build the BDD for a Boolean expression.  There are 3 possibile
 * terminal conditions: 1) the formula evaluates to the constant zero, 2) the formula
 * evaluates to the constant one, or 3) the formula evaluates to a single one of its
 * operands.  In any of these cases, we return TRUE and set "result", but we don't set either
 * new_term_case or new_active_oprnds.  If none of the terminal conditions are met, then
 * we return FALSE and set new_term_case and new_active_oprnds, but not "result".
 *
 * The cur_active_oprnds var_set tells us which operands are in the support of cur_term_case.
 * Based on the values in the array operands, some operands may become inactive. Specifically,
 * those operands that have just become a constant, and those operands whose value no longer
 * matter because some other operands just became constants (e.g. in ((f AND g) OR (h AND k)),
 * if g just became ZERO, then f can be set to an arbitrary value (we choose ONE) and set to
 * inactive.
 */
boolean
bdd_bf_test_terminal_case(mgr, operands, oprnds_cof_info, cur_term_case, new_term_case, 
			  cur_active_oprnds, new_active_oprnds, result)
bdd_manager *mgr;
array_t *operands;              /* bdd_node * */
array_t *oprnds_cof_info; /* bdd_bf_cofactor_info_t * */
bdd_t *cur_term_case;
bdd_t **new_term_case;          /* return; not set if return is TRUE */
var_set_t *cur_active_oprnds;
var_set_t **new_active_oprnds;	/* return; not set if return is TRUE */
bdd_node **result;              /* return; not set if return is TRUE */
{
    bdd_node *oprnd;
    bdd_t *cube, *temp, *function_value;
    int i, j;
    bdd_t *eqn_var;
    array_t *eqn_vars;       /* bdd_t * */
    bfbdd_info_t *bfbdd_info;
    int num_operands, num_eqn_vars, num_vars;
    var_set_t *new_active_oprnds_not, *still_active, *just_became_inactive;
    boolean is_complemented, terminal_case;
    bdd_variableId oprnd_id;
    bdd_bf_cofactor_info_t *cofactor_info;

    /*
     * NOTE: oprnd is not protected for garbage collection.  It's assumed that from the point
     * it is set to the latest point it is read, no operations are performed that would trigger
     * a garbage collection.  ALSO, it's assumed that the array operands has been protected
     * by the calling routine (it's incorrect to protect something more than once).
     */

    bfbdd_info = (bfbdd_info_t *) mgr->undef1;
    eqn_vars = bfbdd_info->eqn_vars;
    num_eqn_vars = array_n(eqn_vars);
    num_operands = array_n(operands);

    /*
     * still_active starts out as cur_active_oprnds.  Any operand which is currently active and
     * is a constant is dropped from still_active.
     */
    still_active = var_set_copy(cur_active_oprnds);

    /*
     * Form a cube giving the value of the eqn_vars for those operands that are currently active
     * and are a constant value.
     */
    cube = bdd_one(bfbdd_info->ev_mgr);  /* this calculation done in the eqn_vars manager */
    for (i = 0; i < num_operands; i++) {
	if (var_set_get_elt(cur_active_oprnds, i)) { 
	    oprnd = array_fetch(bdd_node *, operands, i);
	    if ((oprnd == BDD_ONE(mgr)) || (oprnd == BDD_ZERO(mgr))) {
		var_set_clear_elt(still_active, i); 
		eqn_var = array_fetch(bdd_t *, eqn_vars, i);
		if (oprnd == BDD_ONE(mgr)) {
		    function_value = bdd_dup(eqn_var);
		} else { /* (oprnd == BDD_ZERO(mgr)) */
		    function_value = bdd_not(eqn_var);
		}

		temp = bdd_and(cube, function_value, 1, 1);
		bdd_free(cube);  
		bdd_free(function_value);  
		cube = temp;
	    }
	}
    }

    /*
     * Evaluate the current terminal case on those eqn_vars whose corresponding operands are
     * currently active and are a constant value.  Once an operand becomes a constant, then it
     * remains the same constant every where below this point in the recursion tree. Thus,
     * the result of the cofactor represents the new terminal condition for points below in
     * the recursion.
     */
    *new_term_case = bdd_cofactor(cur_term_case, cube);
    bdd_free(cube);

    /*
     * Based on the valueof new_term_case, see if we have hit a terminal condition.
     */
    if (bdd_is_tautology(*new_term_case, 1)) {
	*result = BDD_ONE(mgr);
	var_set_free(still_active);
	bdd_free(*new_term_case);
	return TRUE;
    } else if (bdd_is_tautology(*new_term_case, 0)) {
	*result = BDD_ZERO(mgr);
	var_set_free(still_active);
	bdd_free(*new_term_case);
	return TRUE;
    } else { 

	/* 
	 * The call to bdd_get_support assumes that varId i corresponds to the ith eqn_var.
	 * TODO: we could replace the above call to bdd_cofactor with a function which computes
	 * the cofactor, and returns the support of the result.
	 * NOTE: since the length of the var_set_t that bdd_get_support returns is the number of
	 * variables in the eqn_vars manager, all the active var_sets must have this length,
	 * so that we can perform operations on them.  If we write our own bdd_cofactor, then
	 * we should be able to get rid of this little nuisance.
         * Note that new_active_operands monotonically approaches all inactive.
	 */
	*new_active_oprnds = bdd_get_support(*new_term_case);

	/*
	 * If there is only one variable in the support of new_term_case, then that means that
	 * we are down to a single active operand at this point in the recursion.  If there
	 * are no cofactors pending on this operand, then we have hit a terminal condition.
	 */
	if (var_set_n_elts(*new_active_oprnds) == 1) {
	    for (i = 0; i < num_operands; i++) {
		if (var_set_get_elt(*new_active_oprnds, i)) {
		    /*
		     * If no cofactor operations are pending on operand i, then we have hit a 
		     * terminal case.  Otherwise, even though there is only one active operand
		     * left, we must continue the recursion.
		     */
		    terminal_case = TRUE;  /* default value */
		    oprnd = array_fetch(bdd_node *, operands, i);
		    cofactor_info = array_fetch(bdd_bf_cofactor_info_t *, oprnds_cof_info, i); 
		    if (!var_set_is_empty(cofactor_info->active)) {
			/* 
			 * NOTE: here we assume that the id and level are the same thing.  So,
			 * just look through cofactor_info->active, starting from the top id
			 * of oprnd through the end of the variable ordering, to see if there
			 * are any remaining active cofactor variables.  TODO: If we are using
			 * dynamic reordering, this assumption is not valid.
			 */
			oprnd_id = BDD_REGULAR(oprnd)->id;
			num_vars = bdd_num_vars(mgr);
			for (j = oprnd_id; j < num_vars; j++) {
			    if (var_set_get_elt(cofactor_info->active, j)) {
				/* a cofactor is pending on variable j, so must continue recursion */
				terminal_case = FALSE;
				break;
			    }
			}
		    }

		    if (terminal_case == TRUE) {
			/*
			 * The phase of oprnd to return is the same as the phase of the remaining
			 * variable in new_term_case.
			 */
			(void) bdd_get_node(*new_term_case, &is_complemented); /* disregard return value */
			if (is_complemented) {
			    *result = BDD_NOT(oprnd);
			} else { 
			    *result = oprnd;
			}
			var_set_free(still_active);
			var_set_free(*new_active_oprnds);
			bdd_free(*new_term_case);
			return TRUE;
		    }
		    break;
		}
	    }
	}
    }

    /* 
     * If we have reached this point, then we have not hit a terminal condition yet.  
     */

#if defined(BDD_DEBUG_BF) /* { */
    /*
     * The new set should be contained in the current set, because an operand can never
     * become active after it has become inactive.  TODO: for debug purposes only.
     */
    cur_active_oprnds_not = var_set_new(num_eqn_vars);
    cur_active_oprnds_not = var_set_not(cur_active_oprnds_not, cur_active_oprnds);
    if (var_set_intersect(cur_active_oprnds_not, *new_active_oprnds)) {
	fail("bdd_bf_test_terminal_case: new not contained in current");
    }
    var_set_free(cur_active_oprnds_not);
#endif /* } */

    /*
     * For those operands that just implicitly became inactive, replace them by the constant ONE in
     * the operands array.  Since these operands will never be examined again, it is arbitrary
     * which value we choose.  However, to avoid unnecessary recursive calls, we use a constant.
     *
     * Note that an operand which is not a constant can become inactive; e.g. in (f1*f2)+(f3*f4),
     * if f1 becomes ZERO, then f2 automatically becomes inactive, because the final value is 
     * independent of f2.  An operand is implicitly inactive if it meets the following condition:
     * it is still_active and it is not new_active.
     */
    new_active_oprnds_not = var_set_new(num_eqn_vars);
    new_active_oprnds_not = var_set_not(new_active_oprnds_not, *new_active_oprnds);
    just_became_inactive = var_set_new(num_eqn_vars);
    just_became_inactive = var_set_and(just_became_inactive, new_active_oprnds_not, still_active);
    if (!var_set_is_empty(just_became_inactive)) {
	for (i = 0; i < num_operands; i++) {
	    if (var_set_get_elt(just_became_inactive, i)) {
		array_insert(bdd_node *, operands, i, BDD_ONE(mgr));
	    }
	}
    }
    var_set_free(new_active_oprnds_not);
    var_set_free(just_became_inactive);
    var_set_free(still_active);

    return FALSE;
}
