#include "util.h"
#include "array.h"
#include "st.h" 

#include "bdd.h"
#include "bdd_int.h"
#include "bdd_bf.h"
#include "bdd_bf_int.h"

static void print_formula_rec();

/*
 * Initialize the manager for building formulas.
 */
void
bdd_bf_initialize_manager(mgr)
bdd_manager *mgr;
{
    bfbdd_info_t *bfbdd_info;
    int i;

    /*
     * If it is already initialized, then fail.
     */
    bfbdd_info = (bfbdd_info_t *)mgr->undef1;
    if (bfbdd_info != NIL(bfbdd_info_t)) {
	fail("bdd_bf_initialize_manager: bfbdd already initialized");
    }

    /*
     * Allocate the bdd_info.
     */
    bfbdd_info = ALLOC(bfbdd_info_t, 1);
    bfbdd_info->eqn_vars = array_alloc(bdd_t *, 0);
    bfbdd_info->ev_mgr = bdd_start(0);
    bfbdd_info->cofactor_info = NIL(array_t);
    bfbdd_info->sort_cache_at_gc = FALSE;

    /*
     * Set up the bf_cache.
     */
    bfbdd_info->bf_cache.on = TRUE; 
    bfbdd_info->bf_cache.invalidate_on_gc = FALSE; 
    bfbdd_info->bf_cache.resize_at = BDD_DFLT_ITE_RESIZE_AT;
    bfbdd_info->bf_cache.max_size = BDD_DFLT_ITE_MAX_SIZE;
    bfbdd_info->bf_cache.nbuckets = BDD_CACHE_INITIAL_SIZE;
    bfbdd_info->bf_cache.nentries = 0;
    bfbdd_info->bf_cache.buckets = ALLOC(bdd_bf_cache_entry_t *, bfbdd_info->bf_cache.nbuckets);
    if (bfbdd_info->bf_cache.buckets == NIL(bdd_bf_cache_entry_t *)) {
	(void) bdd_memfail(NIL(bdd_manager), "bdd_bf_initialize_manager");
    }
    for (i = 0; i < bfbdd_info->bf_cache.nbuckets; i++) {
	bfbdd_info->bf_cache.buckets[i] = NIL(bdd_bf_cache_entry_t);
    }

    /*
     *    Clear out the stats
     */
    bfbdd_info->stats.bf_cache.hits = 0;
    bfbdd_info->stats.bf_cache.misses = 0;
    bfbdd_info->stats.bf_cache.collisions = 0;
    bfbdd_info->stats.bf_cache.inserts = 0;

    bfbdd_info->stats.bf_ops.calls = 0;
    bfbdd_info->stats.bf_ops.returns.trivial = 0;
    bfbdd_info->stats.bf_ops.returns.cached = 0;
    bfbdd_info->stats.bf_ops.returns.full = 0;

    bfbdd_info->stats.memory.bf_cache = 0;

    /*
     * Attach the bfbdd_info to the manager.  In the future, this should be part of the bdd_manager.
     */
    mgr->undef1 = (char *) bfbdd_info;
    return;

}

/*
 * Uninitialize the manager for building formulas.
 */
void
bdd_bf_uninitialize_manager(mgr)
bdd_manager *mgr;
{
    bfbdd_info_t *bfbdd_info;
    int i;
    bdd_bf_cache_entry_t *bf_entry;

    /*
     * If bfbdd_info is not present, then fail.
     */
    bfbdd_info = (bfbdd_info_t *)mgr->undef1;
    if (bfbdd_info == NIL(bfbdd_info_t)) {
	fail("bdd_bf_uninitialize_manager: no bfbdd info is present");
    }

    /*
     * Free up the eqn_vars, and its associated bdd_manager.  We don't explicitly bdd_free
     * the contents of the eqn_vars array, since we are just going to free the manager.
     * However, note that those external pointers (bdd_t's) will be counted as outstanding
     * when the manager is terminated (this is not a memory leak though).
     */
    array_free(bfbdd_info->eqn_vars);
    bdd_end(bfbdd_info->ev_mgr);

    /*
     * Free the bf_cache, and any entries it contains.
     */
    for (i = 0; i < bfbdd_info->bf_cache.nbuckets; i++) {
	bf_entry = bfbdd_info->bf_cache.buckets[i];
	if (bf_entry != NIL(bdd_bf_cache_entry_t)) {
	    FREE(bf_entry);
	}
    }
    FREE(bfbdd_info->bf_cache.buckets);

    /*
     * Note that the cofactor_info is freed by bdd_bf_clear_cofactor_info, which is called
     * from bdd_build_formula_multi.
     */

    FREE(bfbdd_info);
}

/*
 * Print out stats related to the bfbdd operations.
 */
void
bdd_bf_print_stats(mgr, file)
bdd_manager *mgr;
FILE *file;	
{
    bfbdd_info_t *bfbdd_info;
    int total_bf_cache_lookups;
    bdd_stats stats;

    bfbdd_info = (bfbdd_info_t *)mgr->undef1;
    if (bfbdd_info == NIL(bfbdd_info_t)) {
	fail("bdd_bf_print_stats: no bfbdd info is present");
    }

    /* 
     * Make intermediate calculation.
     */
    total_bf_cache_lookups = bfbdd_info->stats.bf_cache.hits + bfbdd_info->stats.bf_cache.misses;

#if defined(BDD_AUTOMATED_STATISTICS_GATHERING) /* { */
	/*
	 *    Write out the stuff in a way which is easy to parse. Also, include some extra info.
	 */
    (void) fprintf(stderr, "\
stats: bf_cache calls: total: %d, trivial: %d (%4.1f%%), cached: %d (%4.1f%%), full: %d (%4.1f%%)\n\
stats: bf_cache lookups: %d, misses: %d (%4.1f%%)\n\
stats: bf_cache insertions: %d, collisions: %d (%4.1f%%)\n\
stats: bf_cache entries: %d, percent of buckets: %4.1f%%\n\
",

    bfbdd_info->stats.bf_ops.calls,
    bfbdd_info->stats.bf_ops.returns.trivial,
    bdd_get_percentage(bfbdd_info->stats.bf_ops.returns.trivial, bfbdd_info->stats.bf_ops.calls),
    bfbdd_info->stats.bf_ops.returns.cached,
    bdd_get_percentage(bfbdd_info->stats.bf_ops.returns.cached, bfbdd_info->stats.bf_ops.calls),
    bfbdd_info->stats.bf_ops.returns.full,
    bdd_get_percentage(bfbdd_info->stats.bf_ops.returns.full, bfbdd_info->stats.bf_ops.calls),

    total_bf_cache_lookups,
    bfbdd_info->stats.bf_cache.misses, 
    bdd_get_percentage(bfbdd_info->stats.bf_cache.misses, total_bf_cache_lookups),

    bfbdd_info->stats.bf_cache.inserts,
    bfbdd_info->stats.bf_cache.collisions, 
    bdd_get_percentage(bfbdd_info->stats.bf_cache.collisions, bfbdd_info->stats.bf_cache.inserts),
    bfbdd_info->bf_cache.nentries,
    bdd_get_percentage(bfbdd_info->bf_cache.nentries, bfbdd_info->bf_cache.nbuckets));

    /*
     * Print out the stats for the manager which maintains the equation variables.
     */
    (void) fprintf(file, "\nStatistics for the equation variables BDD manager:\n");
    bdd_dump_manager_stats(bfbdd_info->ev_mgr);
    
#else /* } else { */
    /*
     *    Just do the pretty thing of printing out all the 
     *    statistics in a nice pretty (e.g. human readable) way.
     */

    (void) fprintf(file, "\
BFBDD Statistics\n\
\n\
BF Cache: \n\
 Total calls:    %8d\n\
    trivial:    %9.1f%%\n\
    cached:     %9.1f%%\n\
    full:       %9.1f%%\n\
 Total lookups:  %8d\n\
    misses:     %9.1f%%\n\
 Total inserts:  %8d\n\
    collisions: %9.1f%%\n\
 Total buckets:  %8d\n\
    entries:    %9.1f%%\n\
\n\
Memory Usage (bytes):\n\
  bf_cache:        %9d\n\
",
    bfbdd_info->stats.bf_ops.calls,
    bdd_get_percentage(bfbdd_info->stats.bf_ops.returns.trivial, bfbdd_info->stats.bf_ops.calls),
    bdd_get_percentage(bfbdd_info->stats.bf_ops.returns.cached, bfbdd_info->stats.bf_ops.calls),
    bdd_get_percentage(bfbdd_info->stats.bf_ops.returns.full, bfbdd_info->stats.bf_ops.calls),

    total_bf_cache_lookups,
    bdd_get_percentage(bfbdd_info->stats.bf_cache.misses, total_bf_cache_lookups),

    bfbdd_info->stats.bf_cache.inserts,
    bdd_get_percentage(bfbdd_info->stats.bf_cache.collisions, bfbdd_info->stats.bf_cache.inserts),

    bfbdd_info->bf_cache.nbuckets,
    bdd_get_percentage(bfbdd_info->bf_cache.nentries, bfbdd_info->bf_cache.nbuckets),

    ((bfbdd_info->bf_cache.nbuckets * sizeof(bdd_bf_cache_entry_t *)) 
      + (bfbdd_info->bf_cache.nentries * sizeof(bdd_bf_cache_entry_t))));

    /*
     * Print out the stats for the manager which maintains the equation variables.
     */
    (void) fprintf(file, "\nStatistics for the equation variables BDD manager:\n");
    bdd_get_stats(bfbdd_info->ev_mgr, &stats);
    bdd_print_stats(stats, file);

#endif /* } */
}

/*
 * Given an array describing a BDD formula using bdd_fml_type_t's, build a tree describing
 * the same formula.  At the leaves, insert the actual BDD operands, and the eqn_vars.
 */
bdd_formula_t *
bdd_bf_construct_formula_tree(mgr, formula_array, operands)
bdd_manager *mgr;
array_t *formula_array;   /* bdd_fml_type_t * */
array_t *operands;        /* bdd_t * */
{
    int i, formula_length, sp, oprnd_index, num_operands;
    array_t *stack;       /* bdd_formula_t * */
    bdd_formula_t *formula_node;
    bdd_fml_type_t formula_type;
    array_t *eqn_vars;       /* bdd_t * */
    bfbdd_info_t *bfbdd_info;

    /*
     * Get the eqn_vars array.  For the ith operand, we use the ith eqn_var.
     */
    bfbdd_info = (bfbdd_info_t *) mgr->undef1;
    eqn_vars = bfbdd_info->eqn_vars;

    /* 
     * Count the number of operands in the formula, and make sure that it matches the
     * length of the operands array.
     */
    formula_length = array_n(formula_array);
    num_operands = 0;
    for (i = 0; i < formula_length; i++) {
	formula_type = array_fetch(bdd_fml_type_t, formula_array, i);
	if (formula_type == BDD_FML_OPRND) {
	    num_operands++;
	}
    }
    if (num_operands != array_n(operands)) {
	fail("construct_formula_tree: length of operands array doesn't match number of operands in formula");
    }
	
    /*
     * Initialize the stack.  The stack can't be deeper than formula length.
     */
    stack = array_alloc(bdd_formula_t *, formula_length);  
    sp = oprnd_index = 0;

    /*
     * Process each entry in the formula_array.
     */
    for (i = 0; i < formula_length; i++) {
	formula_type = array_fetch(bdd_fml_type_t, formula_array, i);
	formula_node = ALLOC(bdd_formula_t, 1);
	formula_node->type = formula_type;

	switch(formula_type) {
	case BDD_FML_OPRND:
	    /*
	     * This is a leaf of the tree.  Insert the BDD for the operand, and the BDD for
	     * the eqn_var corresponding to the operand.
	     */
	    formula_node->data1.operand = array_fetch(bdd_t *, operands, oprnd_index);
	    formula_node->data2.eqn_var = array_fetch(bdd_t *, eqn_vars, oprnd_index); /* reuse eqn_vars */
	    oprnd_index++;  /* increment to get the next operand in the array */
	    array_insert(bdd_formula_t *, stack, sp++, formula_node);
	    break;
	case BDD_FML_AND:
	case BDD_FML_OR:
	case BDD_FML_XOR:
	case BDD_FML_COF:   /* left is what is being cofactored; right is what we are cofactoring by */
	    formula_node->data2.r = array_fetch(bdd_formula_t *, stack, --sp);
	    formula_node->data1.l = array_fetch(bdd_formula_t *, stack, --sp);
	    array_insert(bdd_formula_t *, stack, sp++, formula_node);
	    break;
	case BDD_FML_NOT:
	    formula_node->data1.l = array_fetch(bdd_formula_t *, stack, --sp);
	    formula_node->data2.r = NIL(bdd_formula_t);
	    array_insert(bdd_formula_t *, stack, sp++, formula_node);
	    break;
	default:
	    fail("construct_formula_tree: unknown formula type");
	}
    }

    /*
     * Fail if the stack is not empty.
     */
    if ((sp - 1) != 0) {
	fail("construct_formula_tree: stack not empty");
    }

    /* 
     * Return the top of the stack.
     */
    formula_node = array_fetch(bdd_formula_t *, stack, 0);
    array_free(stack);
    
    return formula_node;
}

/*
 * Print infix formula with two operands.
 */
static void 
print_formula_binop(f1, f2, binop)
bdd_formula_t *f1;
bdd_formula_t *f2;
char *binop;
{
    printf("(");
    print_formula_rec(f1);
    printf(") %s (", binop);
    print_formula_rec(f2);
    printf(")");
    return;
}

/*
 * Recursive step to print a bdd_formula_t to stdout.
 */
static void 
print_formula_rec(formula)
bdd_formula_t *formula;
{
    bdd_fml_type_t formula_type;

    formula_type = formula->type;
    switch(formula_type) {
    case BDD_FML_OPRND:
	printf("f");  /* maybe could be more creative here */
	break;
    case BDD_FML_AND:
	print_formula_binop(formula->data1.l, formula->data2.r, "*");
	break;
    case BDD_FML_OR:
	print_formula_binop(formula->data1.l, formula->data2.r, "+");
	break;
    case BDD_FML_XOR:
	print_formula_binop(formula->data1.l, formula->data2.r, "^");
	break;
    case BDD_FML_COF:
	printf("Cof(");
	print_formula_rec(formula->data1.l);
	printf(", ");
	print_formula_rec(formula->data2.r);
	printf(")");
	break;
    case BDD_FML_NOT:
	printf("!(");
	print_formula_rec(formula->data1.l);
	printf(")");
	break;
    default:
	fail("print_formula_rec: unknown formula type");
    }
    return;
}

/*
 * Print a bdd_formula_t to stdout.  For debugging purposes.
 */
void 
bdd_bf_print_formula(formula)
bdd_formula_t *formula;
{

    if (formula == NIL(bdd_formula_t)) {
	fail("bdd_bf_print_formula: NIL formula");
    }

    print_formula_rec(formula);
    printf(";\n");
    return;
}

/*
 * Free the memory associated with the formula tree.
 */
void
bdd_bf_free_formula(formula)
bdd_formula_t *formula;
{
    switch(formula->type) {
    case BDD_FML_AND:
    case BDD_FML_OR:
    case BDD_FML_XOR:
    case BDD_FML_COF:
	bdd_bf_free_formula(formula->data2.r);
	/* fall through */
    case BDD_FML_NOT:
	bdd_bf_free_formula(formula->data1.l);
	FREE(formula);
	break;
    case BDD_FML_OPRND:
        /* nothing to do: the BDDs here are taken care of elsewhere */
	break;
    default:
	fail("bdd_bf_free_formula: unknown formula type");
    }
}


