/*
 * This file provides a test routine, accessible from the SIS command line, to
 * test the bfbdd code.  The command works by first reading in a SIS network
 * (e.g. using read_blif foo.blif), and then executing _bfbdd_test.
 */

#include "sis.h"
#include "bdd_int.h" /* for mgr->undef1, bdd_memfail */
#include "bdd_bf.h"
#include "bdd_bf_int.h"

/*
 * Build the BDD for formula, using the first num_operands of fn_array as the operands 
 * of the formula.  Build it using binary operators and using the array formulation, and
 * make sure that the results are equal.
 */
static void
test_on_formula(formula, fn_array, num_operands)
array_t *formula;  /* of bdd_fml_type_t * */       
array_t *fn_array; /* of bdd_t * */
int num_operands;
{
    bdd_t *f, *f_binary, *f_array;
    array_t *operands;    /* bdd_t * */
    int i;
    static int call_number = 0;

    /* 
     * The length of the operands argument to bdd_build_formula must be equal
     * to the number of OPRNDs in formula.  So, get the first num_operands out
     * out of fn_array, and put then into a new array called operands.
     */
    operands = array_alloc(bdd_t *, num_operands);
    for (i = 0; i < num_operands; i++) {
	f = array_fetch(bdd_t *, fn_array, i);
	array_insert(bdd_t *, operands, i, f);
    }

    f_binary = bdd_build_formula(formula, operands, BDD_BUILD_BINARY);
    f_array  = bdd_build_formula(formula, operands, BDD_BUILD_MULTI);
    
    if (!bdd_equal(f_binary, f_array)) {
	printf("test_bfbdd: binary and array give different answers, call_number = %d\n", call_number);
	printf("BINARY ANSWER\n");
	bdd_print(f_binary);
	printf("MULTI ANSWER\n");
	bdd_print(f_array);
    }

    bdd_free(f_binary);
    bdd_free(f_array);
    array_free(operands);
    call_number++;
}

/* 
 * Construct a bunch of different formulas to exercise the bfbdd code.
 */
static void
test_bfbdd(mgr, po_array, pi_array)
bdd_manager *mgr;
array_t *po_array; /* of bdd_t * */
array_t *pi_array; /* of bdd_t * */
{
    array_t *formula;      /* bdd_fml_type_t * */
    array_t *fn_array;     /* bdd_t * */
    bdd_t *f, *f_not, *g, *temp1, *temp2;

    /* 
     * We need enough operands to play around with.
     */
    if (array_n(po_array) < 4) {
	(void) printf("test_bfbdd: Insufficient number of operands (po) to perform test\n");
	return; 
    }
    if (array_n(pi_array) < 3) {
	(void) printf("test_bfbdd: Insufficient number of operands (pi) to perform test\n");
	return; 
    }

    /*
     * Initialize the manager for building formulas.
     */
    bdd_bf_initialize_manager(mgr);

    /* 
     * Construct a formula: f  (not expected to be efficient)
     */
    formula = array_alloc(bdd_fml_type_t, 0);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    test_on_formula(formula, po_array, 1);
    array_free(formula);

    /* 
     * Construct a formula: !f  (not expected to be efficient)
     */
    formula = array_alloc(bdd_fml_type_t, 0);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_NOT);
    test_on_formula(formula, po_array, 1);
    array_free(formula);

    /* 
     * Construct a formula: f1*f2  
     */
    formula = array_alloc(bdd_fml_type_t, 0);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_AND);
    test_on_formula(formula, po_array, 2);
    array_free(formula);

    /* 
     * Construct a formula: f1+f2  
     */
    formula = array_alloc(bdd_fml_type_t, 0);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OR);
    test_on_formula(formula, po_array, 2);
    array_free(formula);

    /* 
     * Construct a formula: f1 xor f2 
     */
    formula = array_alloc(bdd_fml_type_t, 0);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_XOR);
    test_on_formula(formula, po_array, 2);
    array_free(formula);

    /* 
     * Construct a formula: cof(f, x)
     */
    fn_array = array_alloc(bdd_t *, 0);
    f = array_fetch(bdd_t *, po_array, 0);
    array_insert_last(bdd_t *, fn_array, f);
    f = array_fetch(bdd_t *, pi_array, 0);
    array_insert_last(bdd_t *, fn_array, f);
    formula = array_alloc(bdd_fml_type_t, 0);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_COF);  
    test_on_formula(formula, fn_array, 2);
    array_free(formula);
    array_free(fn_array);

    /* 
     * Construct a formula: cof( cof(f1, x) + f2, x)
     * Testing cofactoring by same variable, in same phase.
     */
    fn_array = array_alloc(bdd_t *, 0);
    f = array_fetch(bdd_t *, po_array, 0);
    array_insert_last(bdd_t *, fn_array, f);
    f = array_fetch(bdd_t *, pi_array, 0);
    array_insert_last(bdd_t *, fn_array, f);
    f = array_fetch(bdd_t *, po_array, 1);
    array_insert_last(bdd_t *, fn_array, f);
    f = array_fetch(bdd_t *, pi_array, 1);
    array_insert_last(bdd_t *, fn_array, f);
    formula = array_alloc(bdd_fml_type_t, 0);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_COF);  
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OR);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_COF);  
    test_on_formula(formula, fn_array, 4);
    array_free(formula);
    array_free(fn_array);

    /* 
     * Construct a formula: cof( cof(f1, x) xor f2, y), where x = pi0*pi1, y = pi1*!pi2
     * Testing cofactoring by cubes.
     */
    fn_array = array_alloc(bdd_t *, 0);
    f = array_fetch(bdd_t *, po_array, 0);
    array_insert_last(bdd_t *, fn_array, f);
    f = array_fetch(bdd_t *, pi_array, 0);
    g = array_fetch(bdd_t *, pi_array, 1);
    temp1 = bdd_and(f, g, 1, 1);
    array_insert_last(bdd_t *, fn_array, temp1);
    f = array_fetch(bdd_t *, po_array, 1);
    array_insert_last(bdd_t *, fn_array, f);
    f = array_fetch(bdd_t *, pi_array, 1);
    g = array_fetch(bdd_t *, pi_array, 2);
    temp2 = bdd_and(f, g, 1, 0);
    array_insert_last(bdd_t *, fn_array, temp2);
    formula = array_alloc(bdd_fml_type_t, 0);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_COF);  
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_XOR);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_COF);  
    test_on_formula(formula, fn_array, 4);
    array_free(formula);
    array_free(fn_array);
    bdd_free(temp1);  /* can't free these any earlier, because they are present in fn_array */
    bdd_free(temp2);

    /* 
     * Construct a formula: ( !f1 * f2 ) xor !(f3 + f4)
     */
    formula = array_alloc(bdd_fml_type_t, 0);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_NOT);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_AND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OR);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_NOT);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_XOR);
    test_on_formula(formula, po_array, 4);
    array_free(formula);

#if 0
    /* 
     * Construct a formula: f1 xor f2 
     */
    formula = array_alloc(bdd_fml_type_t, 0);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_XOR);
    test_on_formula(formula, po_array, 3);  /* purposefully wrong number of args, to trigger failure */
    array_free(formula);

    /* 
     * Construct a formula: cof(f1, f2+f3)
     * Should crash, since second arg to cofactor is not a single oprnd.
     */
    formula = array_alloc(bdd_fml_type_t, 0);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OR);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_COF);  
    test_on_formula(formula, po_array, 3); 
    array_free(formula);

    /* 
     * Construct a formula: cof( cof(f1, x) + f2, y), where y = !x
     * Testing cofactoring by same variable, in opposite phase; should fail.
     */
    fn_array = array_alloc(bdd_t *, 0);
    f = array_fetch(bdd_t *, po_array, 0);
    array_insert_last(bdd_t *, fn_array, f);
    f = array_fetch(bdd_t *, pi_array, 0);
    array_insert_last(bdd_t *, fn_array, f);
    f = array_fetch(bdd_t *, po_array, 1);
    array_insert_last(bdd_t *, fn_array, f);
    f = array_fetch(bdd_t *, pi_array, 0);
    f_not = bdd_not(f);
    array_insert_last(bdd_t *, fn_array, f_not);
    formula = array_alloc(bdd_fml_type_t, 0);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_COF);  
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OR);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_OPRND);
    array_insert_last(bdd_fml_type_t, formula, BDD_FML_COF);  
    test_on_formula(formula, fn_array, 4);
    array_free(formula);
    array_free(fn_array);
    bdd_free(f_not);

#endif

    /*
     * Print out stats, and then uninitialize the manager for building formulas.
     */
    bdd_bf_print_stats(mgr, stdout);
    bdd_bf_uninitialize_manager(mgr);

}

/*
 *  Command to test the bfbdd package from the command line.
 */
int
com_bfbdd_test(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    lsGen gen;
    node_t *pi, *po;
    bdd_manager *mgr;
    st_table *leaves;
    array_t *roots, *node_list;      /* node_t * */
    int i, num_pi;
    array_t *po_array, *pi_array;    /* bdd_t * */
    int verbose;
    int c;
    bdd_t *f;
    int varid;

    /*
     * Not presently using any command line arguments.
     */
    verbose = 0;
    util_getopt_reset();
    while ((c = util_getopt(argc, argv, "v")) != EOF) {
	switch(c) {
	case 'v':
	    verbose = 1;
	    break;
	default:
	    goto usage;
	}
    }
    if (argc != util_optind) goto usage;

    /* 
     * Create a manager. The current network must have at least one primary input to run the
     * BDD tests.
     */
    num_pi = network_num_pi(*network);
    if (num_pi == 0) {
	(void) fprintf(miserr, "current network doesn't have any primary inputs\n");
	return 1;
    }
    mgr = ntbdd_start_manager(num_pi);

    /*
     * The leaves of the BDD's will be the PI's of the network.
     * The roots for which we want to build BDD's are the PO's of the network.
     */
    leaves = st_init_table(st_ptrcmp, st_ptrhash);
    foreach_primary_input(*network, gen, pi) {
	st_insert(leaves, (char *) pi, (char *) -1);
    }
    roots = array_alloc(node_t *, 0);
    foreach_primary_output(*network, gen, po) {
	array_insert_last(node_t *, roots, po);
    }

    /*
     * Get an ordering for the leaves.
     */
    node_list = order_dfs(roots, leaves, 0);
    array_free(node_list);  /* node_list not needed */

    /*
     * Build an array of the primary output BDDs.
     */
    po_array = array_alloc(bdd_t *, 0);
    for(i = 0; i < array_n(roots); i++) {
	po = array_fetch(node_t *, roots, i);
	f = ntbdd_node_to_bdd(po, mgr, leaves);
	array_insert_last(bdd_t *, po_array, f); 
    }

    /*
     * Build an array of the primary input BDDs.  Leave out any PIs that 
     * are not in the support of some PO.
     */
    pi_array = array_alloc(bdd_t *, 0);
    foreach_primary_input(*network, gen, pi) {
	if (st_lookup_int(leaves, (char *) pi, (int *) &varid)) {
	    if ((int) varid == -1) {
		continue;  /* this just means that pi was not in the support of any of the roots */
	    } else if ((int) varid > num_pi) {
		fail("com_bfbdd: varid is out of range");
	    } else {
		f = ntbdd_node_to_bdd(pi, mgr, leaves);
		array_insert_last(bdd_t *, pi_array, f); 
	    }
	} else {
	    fail("com_bfbdd: pi is not in leaves table");
	}
    }

    /* 
     * Exercise the bfbdd code.
     */
    test_bfbdd(mgr, po_array, pi_array);

    /*
     * Kill manager.
     */
    ntbdd_end_manager(mgr);
	    
    /* 
     * Free everything else.
     */
    st_free_table(leaves);
    array_free(po_array);
    array_free(pi_array);
    array_free(roots);

    return 0;

 usage:
    (void) fprintf(miserr, "usage: _bfbdd_test [-v]\n");
    return 1;
}

/*
 *  We are working as the "test" package.
 */
void
init_test()
{
    com_add_command("_bfbdd_test", com_bfbdd_test, /* changes_network */ 0);
}

/*
 *  called just before the program terminates
 */
void
end_test()
{
}

